@@ -17,7 +17,7 @@ http://www.crummy.com/software/BeautifulSoup/bs4/doc/
 """
 
 __author__ = "Leonard Richardson (leonardr@segfault.org)"
-__version__ = "4.1.0"
+__version__ = "4.1.1"
 __copyright__ = "Copyright (c) 2004-2012 Leonard Richardson"
 __license__ = "MIT"
 
@@ -267,7 +267,7 @@ class BeautifulSoup(Tag):
 
         for i in range(len(self.tagStack) - 1, 0, -1):
             if (name == self.tagStack[i].name
-                and nsprefix == self.tagStack[i].nsprefix == nsprefix):
+                and nsprefix == self.tagStack[i].prefix):
                 numPops = len(self.tagStack) - i
                 break
         if not inclusivePop:
