@@ -1,5 +1,4 @@
 import collections
-import itertools
 import re
 import sys
 import warnings
@@ -34,6 +33,51 @@ class NamespacedAttribute(unicode):
         obj.namespace = namespace
         return obj
 
+class AttributeValueWithCharsetSubstitution(unicode):
+    """A stand-in object for a character encoding specified in HTML."""
+
+class CharsetMetaAttributeValue(AttributeValueWithCharsetSubstitution):
+    """A generic stand-in for the value of a meta tag's 'charset' attribute.
+
+    When Beautiful Soup parses the markup '<meta charset="utf8">', the
+    value of the 'charset' attribute will be one of these objects.
+    """
+
+    def __new__(cls, original_value):
+        obj = unicode.__new__(cls, original_value)
+        obj.original_value = original_value
+        return obj
+
+    def encode(self, encoding):
+        return encoding
+
+
+class ContentMetaAttributeValue(AttributeValueWithCharsetSubstitution):
+    """A generic stand-in for the value of a meta tag's 'content' attribute.
+
+    When Beautiful Soup parses the markup:
+     <meta http-equiv="content-type" content="text/html; charset=utf8">
+
+    The value of the 'content' attribute will be one of these objects.
+    """
+
+    CHARSET_RE = re.compile("((^|;)\s*charset=)([^;]*)", re.M)
+
+    def __new__(cls, original_value):
+        match = cls.CHARSET_RE.search(original_value)
+        if match is None:
+            # No substitution necessary.
+            return unicode.__new__(unicode, original_value)
+
+        obj = unicode.__new__(cls, original_value)
+        obj.original_value = original_value
+        return obj
+
+    def encode(self, encoding):
+        def rewrite(match):
+            return match.group(1) + encoding
+        return self.CHARSET_RE.sub(rewrite, self.original_value)
+
 
 class PageElement(object):
     """Contains the navigational information for some part of the page
@@ -57,11 +101,25 @@ class PageElement(object):
         None : None
         }
 
+    @classmethod
+    def format_string(self, s, formatter='minimal'):
+        """Format the given string using the given formatter."""
+        if not callable(formatter):
+            formatter = self.FORMATTERS.get(
+                formatter, EntitySubstitution.substitute_xml)
+        if formatter is None:
+            output = s
+        else:
+            output = formatter(s)
+        return output
+
     def setup(self, parent=None, previous_element=None):
         """Sets up the initial relations between this element and
         other elements."""
         self.parent = parent
         self.previous_element = previous_element
+        if previous_element is not None:
+            self.previous_element.next_element = self
         self.next_element = None
         self.previous_sibling = None
         self.next_sibling = None
@@ -79,27 +137,25 @@ class PageElement(object):
             raise ValueError("Cannot replace a Tag with its parent.")
         old_parent = self.parent
         my_index = self.parent.index(self)
-        if (hasattr(replace_with, 'parent')
-            and replace_with.parent is self.parent):
-            # We're replacing this element with one of its siblings.
-            if self.parent.index(replace_with) < my_index:
-                # Furthermore, it comes before this element. That
-                # means that when we extract it, the index of this
-                # element will change.
-                my_index -= 1
         self.extract()
         old_parent.insert(my_index, replace_with)
         return self
     replaceWith = replace_with  # BS3
 
-    def replace_with_children(self):
+    def unwrap(self):
         my_parent = self.parent
         my_index = self.parent.index(self)
         self.extract()
         for child in reversed(self.contents[:]):
             my_parent.insert(my_index, child)
         return self
-    replaceWithChildren = replace_with_children  # BS3
+    replace_with_children = unwrap
+    replaceWithChildren = unwrap  # BS3
+
+    def wrap(self, wrap_inside):
+        me = self.replace_with(wrap_inside)
+        wrap_inside.append(me)
+        return wrap_inside
 
     def extract(self):
         """Destructively rips this element out of the tree."""
@@ -148,11 +204,12 @@ class PageElement(object):
             # We're 'inserting' an element that's already one
             # of this object's children.
             if new_child.parent is self:
-                if self.index(new_child) > position:
-                    # Furthermore we're moving it further down the
-                    # list of this object's children. That means that
-                    # when we extract this element, our target index
-                    # will jump down one.
+                current_index = self.index(new_child)
+                if current_index < position:
+                    # We're moving this element further down the list
+                    # of this object's children. That means that when
+                    # we extract this element, our target index will
+                    # jump down one.
                     position -= 1
             new_child.extract()
 
@@ -577,11 +634,6 @@ class PageElement(object):
     def parentGenerator(self):
         return self.parents
 
-    # Utility methods
-    def substitute_encoding(self, str, encoding=None):
-        encoding = encoding or "utf-8"
-        return str.replace("%SOUP-ENCODING%", encoding)
-
 
 class NavigableString(unicode, PageElement):
 
@@ -615,41 +667,45 @@ class NavigableString(unicode, PageElement):
                     self.__class__.__name__, attr))
 
     def output_ready(self, formatter="minimal"):
-        if not callable(formatter):
-            formatter = self.FORMATTERS.get(
-                formatter, EntitySubstitution.substitute_xml)
-        if formatter is None:
-            output = self
-        else:
-            output = formatter(self)
-
+        output = self.format_string(self, formatter)
         return self.PREFIX + output + self.SUFFIX
 
 
-class CData(NavigableString):
+class PreformattedString(NavigableString):
+    """A NavigableString not subject to the normal formatting rules.
+
+    The string will be passed into the formatter (to trigger side effects),
+    but the return value will be ignored.
+    """
+
+    def output_ready(self, formatter="minimal"):
+        """CData strings are passed into the formatter.
+        But the return value is ignored."""
+        self.format_string(self, formatter)
+        return self.PREFIX + self + self.SUFFIX
+
+class CData(PreformattedString):
 
     PREFIX = u'<![CDATA['
     SUFFIX = u']]>'
 
-
-class ProcessingInstruction(NavigableString):
+class ProcessingInstruction(PreformattedString):
 
     PREFIX = u'<?'
     SUFFIX = u'?>'
 
-
-class Comment(NavigableString):
+class Comment(PreformattedString):
 
     PREFIX = u'<!--'
     SUFFIX = u'-->'
 
 
-class Declaration(NavigableString):
+class Declaration(PreformattedString):
     PREFIX = u'<!'
     SUFFIX = u'!>'
 
 
-class Doctype(NavigableString):
+class Doctype(PreformattedString):
 
     @classmethod
     def for_name_and_ids(cls, name, pub_id, system_id):
@@ -688,20 +744,11 @@ class Tag(PageElement):
         self.prefix = prefix
         if attrs is None:
             attrs = {}
+        elif builder.cdata_list_attributes:
+            attrs = builder._replace_cdata_list_attribute_values(
+                self.name, attrs)
         else:
             attrs = dict(attrs)
-            if builder.cdata_list_attributes:
-                universal = builder.cdata_list_attributes.get('*', [])
-                tag_specific = builder.cdata_list_attributes.get(
-                    self.name.lower(), [])
-                for cdata_list_attr in itertools.chain(universal, tag_specific):
-                    if cdata_list_attr in attrs:
-                        # Basically, we have a "class" attribute whose
-                        # value is a whitespace-separated list of CSS
-                        # classes. Split it into a list.
-                        value = attrs[cdata_list_attr]
-                        values = whitespace_re.split(value)
-                        attrs[cdata_list_attr] = values
         self.attrs = attrs
         self.contents = []
         self.setup(parent, previous)
@@ -709,11 +756,9 @@ class Tag(PageElement):
 
         # Set up any substitutions, such as the charset in a META tag.
         if builder is not None:
-            self.contains_substitutions = builder.set_up_substitutions(self)
-
+            builder.set_up_substitutions(self)
             self.can_be_empty_element = builder.can_be_empty_element(name)
         else:
-            self.contains_substitutions = False
             self.can_be_empty_element = False
 
     parserClass = _alias("parser_class")  # BS3
@@ -756,7 +801,7 @@ class Tag(PageElement):
     @string.setter
     def string(self, string):
         self.clear()
-        self.append(string)
+        self.append(string.__class__(string))
 
     def _all_strings(self, strip=False):
         """Yield all child strings, possibly stripping them."""
@@ -775,7 +820,7 @@ class Tag(PageElement):
         for string in self._all_strings(True):
             yield string
 
-    def get_text(self, separator="", strip=False):
+    def get_text(self, separator=u"", strip=False):
         """
         Get all child strings, concatenated using the given separator.
         """
@@ -943,25 +988,28 @@ class Tag(PageElement):
                         val = ' '.join(val)
                     elif not isinstance(val, basestring):
                         val = str(val)
-                    if (self.contains_substitutions
-                        and eventual_encoding is not None
-                        and '%SOUP-ENCODING%' in val):
-                        val = self.substitute_encoding(val, eventual_encoding)
+                    elif (
+                        isinstance(val, AttributeValueWithCharsetSubstitution)
+                        and eventual_encoding is not None):
+                        val = val.encode(eventual_encoding)
 
-                    decoded = (str(key) + '='
-                               + EntitySubstitution.substitute_xml(val, True))
+                    text = self.format_string(val, formatter)
+                    decoded = (
+                        str(key) + '='
+                        + EntitySubstitution.quoted_attribute_value(text))
                 attrs.append(decoded)
         close = ''
         closeTag = ''
-        if self.is_empty_element:
-            close = '/'
-        else:
-            closeTag = '</%s>' % self.name
 
         prefix = ''
         if self.prefix:
             prefix = self.prefix + ":"
 
+        if self.is_empty_element:
+            close = '/'
+        else:
+            closeTag = '</%s%s>' % (prefix, self.name)
+
         pretty_print = (indent_level is not None)
         if pretty_print:
             space = (' ' * (indent_level - 1))
@@ -1034,6 +1082,21 @@ class Tag(PageElement):
                     s.append("\n")
         return ''.join(s)
 
+    def encode_contents(
+        self, indent_level=None, encoding=DEFAULT_OUTPUT_ENCODING,
+        formatter="minimal"):
+        """Renders the contents of this tag as a bytestring."""
+        contents = self.decode_contents(indent_level, encoding, formatter)
+        return contents.encode(encoding)
+
+    # Old method for BS3 compatibility
+    def renderContents(self, encoding=DEFAULT_OUTPUT_ENCODING,
+                       prettyPrint=False, indentLevel=0):
+        if not prettyPrint:
+            indentLevel = None
+        return self.encode_contents(
+            indent_level=indentLevel, encoding=encoding)
+
     #Soup methods
 
     def find(self, name=None, attrs={}, recursive=True, text=None,
@@ -1058,6 +1121,7 @@ class Tag(PageElement):
         callable that takes a string and returns whether or not the
         string matches for some custom definition of 'matches'. The
         same is true of the tag name."""
+
         generator = self.descendants
         if not recursive:
             generator = self.children
@@ -1099,20 +1163,55 @@ class SoupStrainer(object):
     text)."""
 
     def __init__(self, name=None, attrs={}, text=None, **kwargs):
-        self.name = name
+        self.name = self._normalize_search_value(name)
         if not isinstance(attrs, dict):
             # Treat a non-dict value for attrs as a search for the 'class'
             # attribute.
             kwargs['class'] = attrs
             attrs = None
+
         if kwargs:
             if attrs:
                 attrs = attrs.copy()
                 attrs.update(kwargs)
             else:
                 attrs = kwargs
-        self.attrs = attrs
-        self.text = text
+        normalized_attrs = {}
+        for key, value in attrs.items():
+            normalized_attrs[key] = self._normalize_search_value(value)
+
+        self.attrs = normalized_attrs
+        self.text = self._normalize_search_value(text)
+
+    def _normalize_search_value(self, value):
+        # Leave it alone if it's a Unicode string, a callable, a
+        # regular expression, a boolean, or None.
+        if (isinstance(value, unicode) or callable(value) or hasattr(value, 'match')
+            or isinstance(value, bool) or value is None):
+            return value
+
+        # If it's a bytestring, convert it to Unicode, treating it as UTF-8.
+        if isinstance(value, bytes):
+            return value.decode("utf8")
+
+        # If it's listlike, convert it into a list of strings.
+        if hasattr(value, '__iter__'):
+            new_value = []
+            for v in value:
+                if (hasattr(v, '__iter__') and not isinstance(v, bytes)
+                    and not isinstance(v, unicode)):
+                    # This is almost certainly the user's mistake. In the
+                    # interests of avoiding infinite loops, we'll let
+                    # it through as-is rather than doing a recursive call.
+                    new_value.append(v)
+                else:
+                    new_value.append(self._normalize_search_value(v))
+            return new_value
+
+        # Otherwise, convert it into a Unicode string.
+        # The unicode(str()) thing is so this will do the same thing on Python 2
+        # and Python 3.
+        return unicode(str(value))
 
     def __str__(self):
         if self.text:
@@ -1188,13 +1287,12 @@ class SoupStrainer(object):
         return found
 
     def _matches(self, markup, match_against):
-        #print "Matching %s against %s" % (markup, match_against)
+        # print u"Matching %s against %s" % (markup, match_against)
         result = False
-
         if isinstance(markup, list) or isinstance(markup, tuple):
-            # This should only happen when searching, e.g. the 'class'
-            # attribute.
-            if (isinstance(match_against, basestring)
+            # This should only happen when searching a multi-valued attribute
+            # like 'class'.
+            if (isinstance(match_against, unicode)
                 and ' ' in match_against):
                 # A bit of a special case. If they try to match "foo
                 # bar" on a multivalue attribute's value, only accept
@@ -1203,41 +1301,44 @@ class SoupStrainer(object):
                 # XXX This is going to be pretty slow because we keep
                 # splitting match_against. But it shouldn't come up
                 # too often.
-                result = (whitespace_re.split(match_against) == markup)
+                return (whitespace_re.split(match_against) == markup)
             else:
                 for item in markup:
                     if self._matches(item, match_against):
-                        result = True
-        elif match_against is True:
-            result = markup is not None
-        elif isinstance(match_against, collections.Callable):
-            result = match_against(markup)
-        else:
-            #Custom match methods take the tag as an argument, but all
-            #other ways of matching match the tag name as a string.
-            if isinstance(markup, Tag):
-                markup = markup.name
-            if markup is not None and not isinstance(markup, basestring):
-                markup = unicode(markup)
-            #Now we know that chunk is either a string, or None.
-            if hasattr(match_against, 'match'):
-                # It's a regexp object.
-                result = markup and match_against.search(markup)
-            elif (hasattr(match_against, '__iter__')
-                    and markup is not None
-                    and not isinstance(match_against, basestring)):
-                result = markup in match_against
-            elif hasattr(match_against, 'items'):
-                if markup is None:
-                    result = len(match_against.items()) == 0
-                else:
-                    result = match_against in markup
-            elif match_against and isinstance(markup, basestring):
-                match_against = markup.__class__(match_against)
+                        return True
+                return False
 
-            if not result:
-                result = match_against == markup
-        return result
+        if match_against is True:
+            # True matches any non-None value.
+            return markup is not None
+
+        if isinstance(match_against, collections.Callable):
+            return match_against(markup)
+
+        # Custom callables take the tag as an argument, but all
+        # other ways of matching match the tag name as a string.
+        if isinstance(markup, Tag):
+            markup = markup.name
+
+        # Ensure that `markup` is either a Unicode string, or None.
+        markup = self._normalize_search_value(markup)
+
+        if markup is None:
+            # None matches None, False, an empty string, an empty list, and so on.
+            return not match_against
+
+        if isinstance(match_against, unicode):
+            # Exact string match
+            return markup == match_against
+
+        if hasattr(match_against, 'match'):
+            # Regexp match
+            return match_against.search(markup)
+
+        if hasattr(match_against, '__iter__'):
+            # The markup must be an exact match against something
+            # in the iterable.
+            return markup in match_against
 
 
 class ResultSet(list):
